""" 'cnsts.py' is dedicated to the definitions of the main PoPEx constants.
"""

# Tolerance of machine precision
# ------------------------------
NP_MIN_TOL = 1e-12

# Bounds for optimization
# -----------------------
ALPHA_MIN = 0.01
ALPHA_MAX = 100

# Job control
# -----------
NRERUN = 3

# Recompute p_cat frequency
# -------------------------
RECMP_P_CAT_FREQ = 1000